/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 2; tab-width: 2 -*- */
/* gdome-xml-element.c
 *
 * Copyright (C) 1999 Raph Levien <raph@acm.org>
 * Copyright (C) 2000 Mathieu Lacage <mathieu@gnu.org>
 * Copyright (C) 2001 Paolo Casrini <paolo@casarini.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include "gdome.h"
#include "gdome-util.h"
#include <tree.h>
#include <parser.h>
#include <xmlmemory.h>
#include <hash.h>
#include "gdome-xml-util.h"
#include "gdome-xml-str.h"
#include "gdome-xml-node.h"
#include "gdome-xml-nodel.h"
#include "gdome-xml-element.h"
#include "gdome-xml-attribute.h"
#include "gdome-xml-nnodem.h"
#include "gdome-xml-xmlutil.h"


/* vtable for the Element */
const GdomeElementVtab gdome_xml_el_vtab = {
  {
    gdome_xml_n_ref,
    gdome_xml_n_unref,
    gdome_xml_n_query_interface,
    gdome_xml_n_nodeName,
    gdome_xml_n_nodeValue,
    gdome_xml_n_set_nodeValue,
    gdome_xml_n_nodeType,
    gdome_xml_n_parentNode,
    gdome_xml_n_childNodes,
    gdome_xml_n_firstChild,
    gdome_xml_n_lastChild,
    gdome_xml_n_previousSibling,
    gdome_xml_n_nextSibling,
    gdome_xml_el_attributes,
    gdome_xml_n_ownerDocument,
    gdome_xml_n_insertBefore,
    gdome_xml_n_replaceChild,
    gdome_xml_n_removeChild,
    gdome_xml_n_appendChild,
    gdome_xml_n_hasChildNodes,
    gdome_xml_n_cloneNode,
    gdome_xml_n_normalize,
    gdome_xml_n_isSupported,
    gdome_xml_n_namespaceURI,
    gdome_xml_n_prefix,
    gdome_xml_n_set_prefix,
    gdome_xml_n_localName,
    gdome_xml_n_hasAttributes,
    gdome_xml_n_addEventListener,
    gdome_xml_n_removeEventListener,
    gdome_xml_n_dispatchEvent
  },
  gdome_xml_el_tagName,
  gdome_xml_el_getAttribute,
  gdome_xml_el_setAttribute,
  gdome_xml_el_removeAttribute,
  gdome_xml_el_getAttributeNode,
  gdome_xml_el_setAttributeNode,
  gdome_xml_el_removeAttributeNode,
  gdome_xml_el_getElementsByTagName,
  gdome_xml_el_getAttributeNS,
  gdome_xml_el_setAttributeNS,
  gdome_xml_el_removeAttributeNS,
  gdome_xml_el_getAttributeNodeNS,
  gdome_xml_el_setAttributeNodeNS,
  gdome_xml_el_getElementsByTagNameNS,
	gdome_xml_el_hasAttribute,
	gdome_xml_el_hasAttributeNs
};

/**
 * gdome_xml_el_tagName:
 * @self:  Element Objects ref
 * @exc:  Exception Object ref
 *
 * Returns the qualified name of the element.
 */
GdomeDOMString *
gdome_xml_el_tagName (GdomeElement *self, GdomeException *exc) {
	Gdome_xml_Element *priv = (Gdome_xml_Element *)self;
  GdomeDOMString *ret = NULL;
 
  g_return_val_if_fail (priv != NULL, NULL);
  g_return_val_if_fail (GDOME_XML_IS_EL (priv), NULL);
  g_return_val_if_fail (exc != NULL, NULL);

	if(gdome_xmlGetNs(priv->n) != NULL)
		ret = gdome_xml_str_mkref_own (g_strdup_printf("%s:%s",
																									 (gchar *) gdome_xmlGetNsPrefix(priv->n),
																									 (gchar *) gdome_xmlGetName(priv->n)));
	else
		ret = gdome_xml_str_mkref_dup((gchar *)gdome_xmlGetName (priv->n));

	return ret;
}

/**
 * gdome_xml_el_attributes:
 * @self:  Element Objects ref
 * @exc:  Exception Object ref
 *
 * Returns a NamedNodeMap of the element's attrubutes.
 */
GdomeNamedNodeMap *
gdome_xml_el_attributes (GdomeNode *self, GdomeException *exc)
{
	Gdome_xml_Node *priv = (Gdome_xml_Node *)self;

	if (priv == NULL || exc == NULL)
		return NULL;
	g_return_val_if_fail (GDOME_XML_IS_EL (priv), NULL);

  return gdome_xml_nnm_mkref((GdomeDocument *)gdome_xml_n_mkref ((xmlNode *)gdome_xmlGetOwner (priv->n)),
                             (GdomeElement *)self, gdome_xmlGetAttrList(priv->n),
                             priv->accessType, GDOME_XML_NNM_LIST);
}

/**
 * gdome_xml_el_getAttribute:
 * @self:  Element Objects ref
 * @name:  The name of attribute of which the value is wanted
 * @exc:  Exception Object ref
 *
 * Retrieves an attribute value by name.
 * Returns the Attr value as a string, or the empty string if that attribute
 * does not have a specified or default value.
 */
GdomeDOMString *
gdome_xml_el_getAttribute (GdomeElement *self, GdomeDOMString *name, GdomeException *exc)
{
  Gdome_xml_Element *priv = (Gdome_xml_Element *)self;

  if (priv == NULL || name == NULL || exc == NULL)
    return NULL;
  g_return_val_if_fail (GDOME_XML_IS_EL (priv), NULL);

  return gdome_xml_str_mkref_own (xmlGetProp (priv->n, name->str));
}

/**
 * gdome_xml_el_setAttribute:
 * @self:  Element Objects ref
 * @name:  The name of the attribute to create or alter
 * @value: Value to set in string form
 * @exc:  Exception Object ref
 *
 * Adds a new attribute. If an attribute with that name is already present in
 * the element, its value is changed to be that of the value parameter.
 *
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
 */
void
gdome_xml_el_setAttribute (GdomeElement *self, GdomeDOMString *name, GdomeDOMString *value, GdomeException *exc)
{
  Gdome_xml_Element *priv = (Gdome_xml_Element *)self;
  xmlNode *n = NULL;
  CHAR *prevValue = NULL;

  if (priv == NULL) { return ; }
  g_return_if_fail (GDOME_XML_IS_EL (priv));
  if (name == NULL) { return; }
  if (value == NULL) { return; }
  if (exc == NULL) { return; }

  if (GDOME_ISREADONLY (priv)) {
    *exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
    return;
  }

  n = priv->n;
  prevValue = xmlGetProp (n, name->str);

  xmlSetProp (n, name->str, value->str);

  gdome_xml_mut_invoke_listener ("DOMAttrModified",
																 (GdomeNode *)self,
																 TRUE,
																 FALSE,
																 NULL,
																 prevValue,
																 value->str,
																 name->str);
  xmlFree (prevValue);
	
	gdome_xml_mut_invoke_listener ("DOMSubtreeModified",
																 (GdomeNode *)self,
																 TRUE,
																 FALSE,
																 NULL,
																 NULL,
																 NULL,
																 NULL);
}

/**
 * gdome_xml_el_removeAttribute:
 * @self:  Element Objects ref
 * @name:  The name of the attribute to remove
 * @exc:  Exception Object ref
 *
 * Removes an attribute by name. If the removed attribute is known to have a
 * default value, an attribute immediately appears containing the default
 * value as well as the corresponding namespace URI, local name, and prefix
 * when applicable.
 * To remove an attribute by local name and namespace URI, use the
 * removeAttributeNS method.
 *
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
 */
void
gdome_xml_el_removeAttribute (GdomeElement *self, GdomeDOMString *name, GdomeException *exc)
{
  Gdome_xml_Element *priv = (Gdome_xml_Element *)self;
  GdomeNamedNodeMap *nnm = NULL;
	GdomeNode *attr = NULL;
  xmlAttr *xmlattr = NULL;

  if (priv == NULL || name == NULL || exc == NULL)
    return;
  g_return_if_fail (GDOME_XML_IS_EL (priv));

  if (GDOME_ISREADONLY (priv)) {
    *exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
    return;
  }

  nnm =  gdome_xml_nnm_mkref((GdomeDocument *)gdome_xml_n_mkref ((xmlNode *)gdome_xmlGetOwner (priv->n)),
                             self, gdome_xmlGetAttrList(priv->n), priv->accessType, GDOME_XML_NNM_LIST);
	attr = gdome_xml_nnm_removeNamedItem(nnm, name, exc);
  gdome_xml_nnm_unref(nnm, exc);

	xmlattr = (xmlAttr *) ((Gdome_xml_Node*)attr)->n;
	gdome_xml_n_unref(attr, exc);
  xmlFreeProp(xmlattr);
}

/**
 * gdome_xml_el_getAttributeNode:
 * @self:  Element Objects ref
 * @name:  The name of the attribute to retreive
 * @exc:  Exception Object ref
 *
 * Retrieves an attribute node by name.
 * To retrieve an attribute node by qualified name and namespace URI, use the
 * getAttributeNodeNS method.
 * Returns the Attr node with the specified name or null if there is no such
 * attribute.
 */
GdomeAttr *
gdome_xml_el_getAttributeNode (GdomeElement *self, GdomeDOMString *name, GdomeException *exc)
{
	Gdome_xml_Element *priv = (Gdome_xml_Element *)self;
	GdomeNamedNodeMap *nnm;
  GdomeAttr *ret;

	if (priv == NULL || name == NULL || exc == NULL)
		return NULL;
  g_return_val_if_fail (GDOME_XML_IS_EL (priv), NULL);

  nnm = gdome_xml_nnm_mkref((GdomeDocument *)gdome_xml_n_mkref ((xmlNode *)gdome_xmlGetOwner (priv->n)),
                            self, gdome_xmlGetAttrList(priv->n), priv->accessType, GDOME_XML_NNM_LIST);
	ret = (GdomeAttr *)gdome_xml_nnm_getNamedItem(nnm, name, exc);
	gdome_xml_nnm_unref(nnm, exc);

  return ret;
}

/**
 * gdome_xml_el_setAttributeNode:
 * @self:  Element Object ref
 * @newAttr:  The Attr node to add to the attribute list
 * @exc:  Exception Object ref
 *
 * Adds a new attribute node. If an attribute with that name is already present
 * in the element, it is replaced by the new one.
 * To add a new attribute node with a qualified name and namespace URI, use the
 * setAttributeNodeNS method.
 * Returns if the newAttr attribute replaces an existing attribute, the
 * replaced Attr node is returned, otherwise null is returned.
 *
 * GDOME_WRONG_DOCUMENT_ERR: Raised if newAttr was created from a different
 * document than the one that created the element.
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
 * GDOME_INUSE_ATTRIBUTE_ERR: Raised if newAttr is already an attribute of
 * another Element object. The DOM user must explicitly clone Attr nodes to
 * re-use them in other elements.
 */
GdomeAttr *
gdome_xml_el_setAttributeNode (GdomeElement *self, GdomeAttr *newAttr, GdomeException *exc)
{
  Gdome_xml_Element *priv = (Gdome_xml_Element *)self;
	GdomeNamedNodeMap *nnm;
	GdomeAttr *ret;

  if (priv == NULL || newAttr == NULL || exc == NULL)
		return NULL;
	g_return_val_if_fail(GDOME_XML_IS_EL (priv), NULL);

  if (GDOME_ISREADONLY (priv)) {
    *exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
    return NULL;
  }

  nnm =  gdome_xml_nnm_mkref((GdomeDocument *)gdome_xml_n_mkref ((xmlNode *)gdome_xmlGetOwner (priv->n)),
                             self, gdome_xmlGetAttrList(priv->n), priv->accessType, GDOME_XML_NNM_LIST);
	ret = (GdomeAttr *)gdome_xml_nnm_setNamedItem(nnm, (GdomeNode *)newAttr, exc);
	gdome_xml_nnm_unref(nnm, exc);
	
	return ret;
}

/**
 * gdome_xml_el_removeAttributeNode:
 * @self:  Element Objects ref
 * @oldAttr:  The Attr node to remove from the attribute list
 * @exc:  Exception Object ref
 *
 * Removes the specified attribute node. If the removed Attr has a default
 * value it is immediately replaced. The replacing attribute has the same
 * namespace URI and local name, as well as the original prefix, when
 * applicable.
 * Returns the Attr node that was removed.
 *
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
 * GDOME_NOT_FOUND_ERR: Raised if oldAttr is not an attribute of the element.
 */
GdomeAttr *
gdome_xml_el_removeAttributeNode (GdomeElement *self, GdomeAttr *oldAttr, GdomeException *exc)
{
  Gdome_xml_Element *priv = (Gdome_xml_Element *)self;
  GdomeDOMString *localName, *namespaceURI;
  GdomeNamedNodeMap *nnm = NULL;
  GdomeAttr *ret = NULL;
  xmlAttr *attr = NULL;

  if (priv == NULL || oldAttr == NULL || exc == NULL)
    return NULL;
  g_return_val_if_fail (GDOME_XML_IS_EL (priv), NULL);

  if (GDOME_ISREADONLY (priv)) {
    *exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
    return NULL;
  }

  attr = (xmlAttr *)((Gdome_xml_Attr *)oldAttr)->n;
  nnm = gdome_xml_nnm_mkref((GdomeDocument *)gdome_xml_n_mkref ((xmlNode *)gdome_xmlGetOwner (priv->n)),
                            self, gdome_xmlGetAttrList(priv->n), priv->accessType, GDOME_XML_NNM_LIST);

  if (attr->ns == NULL) {
    localName = gdome_xml_str_mkref(gdome_xmlGetName((xmlNode *)attr));
    ret = (GdomeAttr *)gdome_xml_nnm_removeNamedItem(nnm, localName, exc);
    localName->unref(localName);
  } else {
    localName = gdome_xml_str_mkref(gdome_xmlGetName((xmlNode *)attr));
    namespaceURI = gdome_xml_str_mkref(gdome_xmlGetNsURI((xmlNode *)attr));
    ret = (GdomeAttr *)gdome_xml_nnm_removeNamedItemNS(nnm, namespaceURI, localName, exc);
    localName->unref(localName);
    namespaceURI->unref(namespaceURI);
  }
  return ret;
}

/**
 * gdome_xml_el_getElementsByTagName:
 * @self:  Element Objects ref
 * @name:  The name of the tag to match on. The special value "*" matches all
 *         tags.
 * @exc:  Exception Object ref
 *
 * Returns a NodeList of all descendant Elements with a given tag name, in the
 * order in which they are encountered in a preorder traversal of this Element tree.
 */
GdomeNodeList *
gdome_xml_el_getElementsByTagName (GdomeElement *self, GdomeDOMString *name, GdomeException *exc)
{
  Gdome_xml_Element *priv = (Gdome_xml_Element *)self;

  if (priv == NULL || name == NULL || exc == NULL)
    return NULL;
  g_return_val_if_fail (GDOME_XML_IS_EL (priv), NULL);

  return gdome_xml_nl_mkref ((GdomeNode*)self, name, NULL, GDOME_READWRITE_NODE);
}

/**
 * gdome_xml_el_getElementsByTagNameNS:
 * @self:  Element Objects ref
 * @namespace:  The namespace URI of the elements to match on. The special
 *              value "*" matches all namespaces.
 * @localName:  The local name of the elements to match on. The special value
 *              "*" matches all local names.
 * @exc:  Exception Object ref
 *
 * Returns a NodeList of all the descendant Elements with a given local name
 * and namespace URI in the order in which they are encountered in a preorder
 * traversal of this Element tree.
 */
GdomeNodeList *
gdome_xml_el_getElementsByTagNameNS (GdomeElement *self, GdomeDOMString *namespaceURI, GdomeDOMString *localName, GdomeException *exc)
{
  Gdome_xml_Element *priv = (Gdome_xml_Element *)self;

  if (priv == NULL || namespaceURI == NULL || localName == NULL || exc == NULL)
    return NULL;
  g_return_val_if_fail (GDOME_XML_IS_EL (priv), NULL);

  return gdome_xml_nl_mkref ((GdomeNode*)self, localName, namespaceURI, GDOME_READWRITE_NODE);
}

/**
 * gdome_xml_el_getAttributeNS:
 * @self:  Element Objects ref
 * @namespaceURI:  The namespaceURI of attribute to retrieve
 * @localName:  The localName of the attribute to retrieve
 * @exc:  Exception Object ref
 *
 * Retrieves an attribute value by local name and namespace URI.
 * Returns the Attr value as a string, or the empty string if that attribute
 * does not have a specified or default value.
 */
GdomeDOMString *
gdome_xml_el_getAttributeNS (GdomeElement *self, GdomeDOMString *namespaceURI, GdomeDOMString *localName, GdomeException *exc)
{
  Gdome_xml_Element *priv = (Gdome_xml_Element *)self;

  if (priv == NULL || namespaceURI == NULL || localName == NULL || exc == NULL)
    return NULL;
  g_return_val_if_fail (GDOME_XML_IS_EL (priv), NULL);

  return gdome_xml_str_mkref_own (xmlGetNsProp (priv->n, localName->str, namespaceURI->str));
}

/**
 * gdome_xml_el_setAttributeNS:
 * @self:  Element Objects ref
 * @namespaceURI:  The namespaceURI of attribute to create or alter
 * @qualifiedName:  The localName of the attribute to create or alter
 * @value: Value to set in string form
 * @exc:  Exception Object ref
 *
 * Adds a new attribute. If an attribute with the same namespaceURI and
 * localName is already present in the element, its value is changed to be
 * that of the value parameter.
 *
 * GDOME_NAMESPACE_ERR: Raised if the qualifiedName is malformed, if the
 * qualifiedName has a prefix and the namespaceURI is null, if the
 * qualifiedName has a prefix that is "xml" and the namespaceURI is different
 * from "http://www.w3.org/XML/1998/namespace", or if the qualifiedName is
 * "xmlns" and the namespaceURI is different from "http://www.w3.org/2000/xmlns/".
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
 */
void
gdome_xml_el_setAttributeNS (GdomeElement *self, GdomeDOMString *namespaceURI, GdomeDOMString *qualifiedName, GdomeDOMString *value, GdomeException *exc)
{
  Gdome_xml_Element *priv = (Gdome_xml_Element *)self;
  gchar **strs;
  gchar *prefix = NULL;
  gchar *localName = NULL;
	xmlNs *ns = NULL;

  if (priv == NULL || qualifiedName == NULL || value == NULL || exc == NULL)
    return;
  g_return_if_fail (GDOME_XML_IS_EL (priv));

  if (GDOME_ISREADONLY (priv)) {
    *exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
    return;
  }

   /* test malformed qualifiedName */
	strs = g_strsplit((gchar *)qualifiedName->str, ":", 0);
  if (strs[0] && strs[1] && !strs[2]) {              /* there's only one ':' */
    if (namespaceURI == NULL) {
      *exc = GDOME_NAMESPACE_ERR;
    } else {
      prefix = g_strdup(strs[0]);
      localName = g_strdup(strs[1]);
      if ((!strcmp (prefix, "xml") && strcmp(namespaceURI->str, "http://www.w3.org/XML/1998/namespace"))) {
        *exc = GDOME_NAMESPACE_ERR;
      }
    }
  } else if (!strs[1] && namespaceURI != NULL) {     /* there's no ':' */
    prefix = g_strdup("");
    localName = g_strdup(strs[0]);
    if (strcmp (localName, "xmlns"))
      *exc = GDOME_NAMESPACE_ERR;
  } else if (strs[0] && strs[1] && strs[2]) {        /* there are more than one ':' */
    *exc = GDOME_NAMESPACE_ERR;
	}
	g_strfreev(strs);

  if (!*exc) {
    if (localName != NULL) {
      ns = xmlSearchNsByHref (gdome_xmlGetOwner(priv->n), priv->n, (xmlChar *)namespaceURI->str);
      if(ns == NULL || !xmlStrEqual((xmlChar *)prefix, ns->prefix))
        ns = xmlNewNs(NULL, (xmlChar *)namespaceURI->str, (xmlChar *)prefix);
      xmlSetNsProp(priv->n, ns, (xmlChar *)localName, (xmlChar *)value->str);
      g_free(prefix);
      g_free(localName);
    } else
      xmlSetProp (priv->n, (xmlChar *)qualifiedName->str, (xmlChar *)value->str);
  } else {
    if (localName != NULL) {
      g_free (prefix);
      g_free (localName);
    }

    return;
  }
}

/**
 * gdome_xml_el_removeAttributeNS:
 * @self:  Element Objects ref
 * @namespace:  The namespaceURI of attribute to remove
 * @localName:  The localName of the attribute to remove
 * @exc:  Exception Object ref
 *
 * Removes an attribute by local name and namespace URI. If the removed
 * attribute has a default value it is immediately replaced. The replacing
 * attribute has the same namespace URI and local name, as well as the
 * original prefix.
 *
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
  */
void
gdome_xml_el_removeAttributeNS (GdomeElement *self, GdomeDOMString *namespaceURI, GdomeDOMString *localName, GdomeException *exc)
{
  Gdome_xml_Element *priv = (Gdome_xml_Element *)self;
	GdomeNamedNodeMap *nnm = NULL;
	GdomeNode *attr = NULL;
  xmlAttr *xmlattr = NULL;

  if (priv == NULL || namespaceURI == NULL || localName == NULL || exc == NULL)
    return;
  g_return_if_fail (GDOME_XML_IS_EL (priv));

  if (GDOME_ISREADONLY (priv)) {
    *exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
    return;
  }

  nnm =  gdome_xml_nnm_mkref((GdomeDocument *)gdome_xml_n_mkref ((xmlNode *)gdome_xmlGetOwner (priv->n)),
                             self, gdome_xmlGetAttrList(priv->n), priv->accessType, GDOME_XML_NNM_LIST);
	attr = (GdomeNode *)gdome_xml_nnm_removeNamedItemNS(nnm, namespaceURI, localName, exc);
  gdome_xml_nnm_unref(nnm, exc);

	xmlattr = (xmlAttr *) ((Gdome_xml_Node*)attr)->n;
	gdome_xml_n_unref(attr, exc);
  xmlFreeProp(xmlattr);

	return;
}

/**
 * gdome_xml_el_getAttributeNodeNS:
 * @self:  Element Objects ref
 * @namespaceURI:  The namespace URI of the attribute to retrieve.
 * @localName:  The local name of the attribute to retrieve.
 * @exc:  Exception Object ref
 *
 * Retrieves an Attr node by local name and namespace URI.
 * The Attr node with the specified attribute local name and namespace URI
 * or null if there is no such attribute.
 */
GdomeAttr *
gdome_xml_el_getAttributeNodeNS (GdomeElement *self, GdomeDOMString *namespaceURI, GdomeDOMString *localName, GdomeException *exc)
{
	Gdome_xml_Element *priv = (Gdome_xml_Element *)self;
	GdomeNamedNodeMap *nnm;
  GdomeAttr *ret;

	if (priv == NULL || namespaceURI == NULL || localName == NULL || exc == NULL)
		return NULL;
  g_return_val_if_fail (GDOME_XML_IS_EL (priv), NULL);

  nnm = gdome_xml_nnm_mkref((GdomeDocument *)gdome_xml_n_mkref ((xmlNode *)gdome_xmlGetOwner (priv->n)),
                            self, gdome_xmlGetAttrList(priv->n), priv->accessType, GDOME_XML_NNM_LIST);
	ret = (GdomeAttr *)gdome_xml_nnm_getNamedItemNS(nnm, namespaceURI, localName, exc);
	gdome_xml_nnm_unref(nnm, exc);

  return ret;
}

/**
 * gdome_xml_el_setAttributeNodeNS:
 * @self:  Element Object ref
 * @newAttr:  The Attr node to add to the attribute list
 * @exc:  Exception Object ref
 *
 * Adds a new attribute. If an attribute with that local name and that
 * namespace URI is already present in the element, it is replaced by the new
 * one.
 * Returns if the newAttr attribute replaces an existing attribute with the
 * same local name and namespace URI, the replaced Attr node is returned,
 * otherwise null is returned.
 *
 * GDOME_WRONG_DOCUMENT_ERR: Raised if newAttr was created from a different
 * document than the one that created the element.
 * GDOME_NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
 * GDOME_INUSE_ATTRIBUTE_ERR: Raised if newAttr is already an attribute of
 * another Element object. The DOM user must explicitly clone Attr nodes to
 * re-use them in other elements.
 */
GdomeAttr *
gdome_xml_el_setAttributeNodeNS (GdomeElement *self, GdomeAttr *newAttr, GdomeException *exc)
{
	Gdome_xml_Element *priv = (Gdome_xml_Element *)self;
	GdomeNamedNodeMap *nnm;
	GdomeAttr *ret;

	if (priv == NULL || newAttr == NULL || exc == NULL)
		return NULL;
	g_return_val_if_fail(GDOME_XML_IS_EL (priv), NULL);

  if (GDOME_ISREADONLY (priv)) {
    *exc = GDOME_NO_MODIFICATION_ALLOWED_ERR;
    return NULL;
  }

  nnm =  gdome_xml_nnm_mkref((GdomeDocument *)gdome_xml_n_mkref ((xmlNode *)gdome_xmlGetOwner (priv->n)),
                             self, gdome_xmlGetAttrList(priv->n), priv->accessType, GDOME_XML_NNM_LIST);
	ret = (GdomeAttr *)gdome_xml_nnm_setNamedItemNS(nnm, (GdomeNode *)newAttr, exc);
	gdome_xml_nnm_unref(nnm, exc);

	return ret;
}

/**
 * gdome_xml_el_hasAttribute:
 * @self:  Element Objects ref
 * @name:  The name of the attribute to look for.
 * @exc:  Exception Object ref
 *
 * Returns true when an attribute with a given name is specified on this
 * element or has a default value, false otherwise.
 */
GdomeBoolean
gdome_xml_el_hasAttribute (GdomeElement *self, GdomeDOMString *name, GdomeException *exc)
{
  Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	xmlChar *val = NULL;

  if (priv == NULL || name == NULL || exc == NULL)
      return 0;
  g_return_val_if_fail (GDOME_XML_IS_EL (priv), 0);

  val = xmlGetProp (priv->n, name->str);

  if (val != NULL) {
    xmlFree(val);
		return 1;
	}

	return 0;
}

/**
 * gdome_xml_el_getAttribute:
 * @self:  Element Objects ref
 * @namespaceURI:  The namespaceURI of the attribute to look for.
 * @localName:  The localName of the attribute to look for.
 * @exc:  Exception Object ref
 *
 * Returns true when an attribute with a given local name and namespace URI
 * is specified on this element or has a default value, false otherwise.
 */
GdomeBoolean         
gdome_xml_el_hasAttributeNs (GdomeElement *self, GdomeDOMString *namespaceURI, GdomeDOMString *localName, GdomeException *exc)
{
  Gdome_xml_Node *priv = (Gdome_xml_Node *)self;
	xmlChar *val;

	if (priv == NULL || namespaceURI == NULL ||
			localName == NULL || exc == NULL)
		return 0;
  g_return_val_if_fail (GDOME_XML_IS_EL (priv), 0);

	val = xmlGetNsProp (priv->n, localName->str, namespaceURI->str);

	if (val != NULL) {
    xmlFree(val);
		return 1;
	}

	return 0;
}

